/* -*-C-*- strtou.c */
/*-->strtoul*/
/**********************************************************************/
/***************************** strtoul ********************************/
/**********************************************************************/

#include "headers.h"

#if defined(HAVE_STRTOUL)
static int have_strtoul = 0;		/* ensure non-empty object file */
					/* because some linkers reject them */
#else

RCSID("$Id: strtou.c,v 1.1 1997/09/01 22:26:35 beebe Exp beebe $")

#define IN(l,a,r) (((l) <= (a)) && ((a) <= (r)))
#define CHECK_OVERFLOW(n,nbits) {if \
	((((n) << (nbits)) >> (nbits)) != (n)) overflow++;}

/* This is a simple implementation of Standard C strtoul().  A library
version should be programmed with more care. */

unsigned long int
#if STDC
strtoul(
const char		*nptr,
char			**endptr,
int			base
)
#else /* NOT STDC */
strtoul(nptr,endptr,base)
const char		*nptr;
char			**endptr;
int			base;
#endif /* STDC */
{
    int			c;		/* current character value */
    int			digit;		/* digit value */
    static const char	*digits = "0123456789abcdefghijklmnopqrstuvxwyz";
    int			has_negsign;	/* non-zero if number negative */
    unsigned long int	number;		/* the accumulating number */
    unsigned long int	old_number;	/* the previous accumulating number */
    int			overflow;
    const char		*pos;		/* pointer into digit list */
    const char		*q;		/* pointer past end of digits */

    overflow = 0;

    if (!(IN(2,base,36) || (base == 0)))
    {
	if (endptr != (char**)NULL)
	    *endptr = (char*)nptr;
	return ((unsigned long int)0);
    }

    while (isspace(*nptr))
	nptr++;				/* ignore leading whitespace */

    q = (char*)nptr;
    if (base == 0)			/* variable base; set by lookahead */
    {
	if (*q == '0')
	    base = ((*(q+1) == 'x') || (*(q+1) == 'X')) ? 16 : 8;
	else
	    base = 10;
    }

    for (has_negsign = 0; *q == '-'; ++q)
	has_negsign = !has_negsign;

    /* eliminate optional "0x" or "0X" prefix */

    if (    (base == 16) &&
	(*q == '0') &&
	((*(q+1) == 'x') || (*(q+1) == 'X')) )
	q += 2;

    number = 0L;

    /* Number conversion is done by shifting rather than multiplication
       when the base is a power of 2, in order that the results not be
       impacted by integer overflow. */
    switch (base)
    {
    case 2:
	while (IN('0',*q,'1'))
	{
	    CHECK_OVERFLOW(number,1);
	    number <<= 1;
	    number |= *q - '0';
	    q++;
	}
	break;

    case 4:
	while (IN('0',*q,'3'))
	{
	    CHECK_OVERFLOW(number,2);
	    number <<= 2;
	    number |= *q - '0';
	    q++;
	}
	break;


    case 8:
	while (IN('0',*q,'7'))
	{
	    CHECK_OVERFLOW(number,3);
	    number <<= 3;
	    number |= *q - '0';
	    q++;
	}
	break;


    case 16:
	for (;;)
	{
	    if (!*q)
		break;
	    c = (unsigned)*q;
	    if (isupper(c))
		c = tolower(c);
	    pos = (const char*)strchr(digits,c);
	    if (pos == (const char*)NULL)
		break;
	    digit = (int)(pos - digits);
	    if (!IN(0,digit,15))
		break;
	    CHECK_OVERFLOW(number,4);
	    number <<= 4;
	    number |= digit;
	    q++;
	}
	break;


    case 32:
	for (;;)
	{
	    if (!*q)
		break;
	    c = (unsigned)*q;
	    if (isupper(c))
		c = tolower(c);
	    pos = (const char*)strchr(digits,c);
	    if (pos == (const char*)NULL)
		break;
	    digit = (int)(pos - digits);
	    if (!IN(0,digit,31))
		break;
	    CHECK_OVERFLOW(number,5);
	    number <<= 5;
	    number |= digit;
	    q++;
	}
	break;

    default:		/* all other bases done by multiplication */
	for (;;)
	{
	    if (!*q)
		break;
	    c = (unsigned)*q;
	    if (isupper(c))
		c = tolower(c);
	    pos = (const char*)strchr(digits,c);
	    if (pos == (const char*)NULL)
		break;
	    digit = (int)(pos - digits);
	    if (!IN(0,digit,base-1))
		break;
	    old_number = number;
	    number *= base;
	    number += digit;
	    if (number < old_number)	/* overflow wraps to smaller value */
		overflow++;
	    q++;
	}
	if (endptr != (char**)NULL)
	    *endptr = (char*)q;
	break;
    }
    if (endptr != (char**)NULL)
	*endptr = (char*)q;
    if (has_negsign)
	number = -number;
    if (overflow)
    {
	number = (unsigned long)ULONG_MAX;
	errno = ERANGE;
    }
    return (number);
}
#endif /* defined(HAVE_STRTOUL) */

#ifdef TEST
/***********************************************************************
Simple test program for strtoul().   Values are read from stdin, and the
results in different bases are echoed to stdout.
***********************************************************************/

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif /* EXIT_SUCCESS */

int
#if STDC
main(
int			argc,
char			*argv[]
)
#else /* NOT STDC */
main(argc,argv)
int			argc;
char			*argv[];
#endif /* STDC */
{
    int base;
    int lineno = 0;
    int n;
    int save_errno;
    unsigned long ul;
    char value[40+1];

    while ((n = scanf("%40s%d",value,&base)) != EOF)
    {
	lineno++;
	if (n == 2)
	{
	    errno = 0;
	    ul = strtoul(value, (char**)NULL, base);
	    save_errno = errno;
	    (void)printf("strtoul(\"%s\",NULL,%d)\t-> 0x%08lx (10#%lu)\n",
			 value, base, ul, ul);
	    if (save_errno != 0)
	    {
		(void)fflush(stdout);
		(void)fprintf(stderr,"errno = %d\t", save_errno);
		(void)perror("perror() says");
		(void)fprintf(stderr,"\n");
		(void)fflush(stderr);
	    }
	}
	else
	    (void)fprintf(stderr,"?Conversion error at line %d\n",lineno);
    }
    exit (EXIT_SUCCESS);
    return (EXIT_SUCCESS);
}
#endif /* TEST */
